@@ -1,21 +1,32 @@
 """Coverage controllers for use by pytest-cov and nose-cov."""
 
+import argparse
 import contextlib
 import copy
 import functools
 import os
 import random
+import shutil
 import socket
 import sys
+import warnings
 from io import StringIO
 from pathlib import Path
+from typing import Union
 
 import coverage
 from coverage.data import CoverageData
+from coverage.sqldata import filename_suffix
 
+from . import CentralCovContextWarning
+from . import DistCovError
 from .embed import cleanup
 
 
+class BrokenCovConfigError(Exception):
+    pass
+
+
 class _NullFile:
     @staticmethod
     def write(v):
@@ -51,16 +62,21 @@ def _ensure_topdir(meth):
     return ensure_topdir_wrapper
 
 
+def _data_suffix(name):
+    return f'{filename_suffix(True)}.{name}'
+
+
 class CovController:
     """Base class for different plugin implementations."""
 
-    def __init__(self, cov_source, cov_report, cov_config, cov_append, cov_branch, config=None, nodeid=None):
+    def __init__(self, options: argparse.Namespace, config: Union[None, object], nodeid: Union[None, str]):
         """Get some common config used by multiple derived classes."""
-        self.cov_source = cov_source
-        self.cov_report = cov_report
-        self.cov_config = cov_config
-        self.cov_append = cov_append
-        self.cov_branch = cov_branch
+        self.cov_source = options.cov_source
+        self.cov_report = options.cov_report
+        self.cov_config = options.cov_config
+        self.cov_append = options.cov_append
+        self.cov_branch = options.cov_branch
+        self.cov_precision = options.cov_precision
         self.config = config
         self.nodeid = nodeid
 
@@ -71,6 +87,7 @@ class CovController:
         self.failed_workers = []
         self.topdir = os.fspath(Path.cwd())
         self.is_collocated = None
+        self.started = False
 
     @contextlib.contextmanager
     def ensure_topdir(self):
@@ -81,6 +98,7 @@ class CovController:
 
     @_ensure_topdir
     def pause(self):
+        self.started = False
         self.cov.stop()
         self.unset_env()
 
@@ -88,6 +106,13 @@ class CovController:
     def resume(self):
         self.cov.start()
         self.set_env()
+        self.started = True
+
+    def start(self):
+        self.started = True
+
+    def finish(self):
+        self.started = False
 
     @_ensure_topdir
     def set_env(self):
@@ -122,15 +147,39 @@ class CovController:
         return 'platform {}, python {}'.format(platform, '{}.{}.{}-{}-{}'.format(*version_info[:5]))
 
     @staticmethod
-    def sep(stream, s, txt):
+    def get_width():
+        # taken from https://github.com/pytest-dev/pytest/blob/33c7b05a/src/_pytest/_io/terminalwriter.py#L26
+        width, _ = shutil.get_terminal_size(fallback=(80, 24))
+        # The Windows get_terminal_size may be bogus, let's sanify a bit.
+        if width < 40:
+            width = 80
+        return width
+
+    def sep(self, stream, s, txt):
         if hasattr(stream, 'sep'):
             stream.sep(s, txt)
         else:
-            sep_total = max((70 - 2 - len(txt)), 2)
-            sep_len = sep_total // 2
-            sep_extra = sep_total % 2
-            out = f'{s * sep_len} {txt} {s * (sep_len + sep_extra)}\n'
-            stream.write(out)
+            fullwidth = self.get_width()
+            # taken from https://github.com/pytest-dev/pytest/blob/33c7b05a/src/_pytest/_io/terminalwriter.py#L126
+            # The goal is to have the line be as long as possible
+            # under the condition that len(line) <= fullwidth.
+            if sys.platform == 'win32':
+                # If we print in the last column on windows we are on a
+                # new line but there is no way to verify/neutralize this
+                # (we may not know the exact line width).
+                # So let's be defensive to avoid empty lines in the output.
+                fullwidth -= 1
+            N = max((fullwidth - len(txt) - 2) // (2 * len(s)), 1)
+            fill = s * N
+            line = f'{fill} {txt} {fill}'
+            # In some situations there is room for an extra sepchar at the right,
+            # in particular if we consider that with a sepchar like "_ " the
+            # trailing space is not important at the end of the line.
+            if len(line) + len(s.rstrip()) <= fullwidth:
+                line += s.rstrip()
+            # (end of terminalwriter borrowed code)
+            line += '\n\n'
+            stream.write(line)
 
     @_ensure_topdir
     def summary(self, stream):
@@ -143,15 +192,15 @@ class CovController:
 
         # Output coverage section header.
         if len(self.node_descs) == 1:
-            self.sep(stream, '-', f"coverage: {''.join(self.node_descs)}")
+            self.sep(stream, '_', f'coverage: {"".join(self.node_descs)}')
         else:
-            self.sep(stream, '-', 'coverage')
+            self.sep(stream, '_', 'coverage')
             for node_desc in sorted(self.node_descs):
                 self.sep(stream, ' ', f'{node_desc}')
 
         # Report on any failed workers.
         if self.failed_workers:
-            self.sep(stream, '-', 'coverage: failed workers')
+            self.sep(stream, '_', 'coverage: failed workers')
             stream.write('The following workers failed to return coverage data, ensure that pytest-cov is installed on these workers.\n')
             for node in self.failed_workers:
                 stream.write(f'{node.gateway.id}\n')
@@ -162,6 +211,7 @@ class CovController:
                 'show_missing': ('term-missing' in self.cov_report) or None,
                 'ignore_errors': True,
                 'file': stream,
+                'precision': self.cov_precision,
             }
             skip_covered = isinstance(self.cov_report, dict) and 'skip-covered' in self.cov_report.values()
             options.update({'skip_covered': skip_covered or None})
@@ -205,6 +255,22 @@ class CovController:
                 total = self.cov.json_report(ignore_errors=True, outfile=output)
             stream.write('Coverage JSON written to file %s\n' % (self.cov.config.json_output if output is None else output))
 
+        # Produce Markdown report if wanted.
+        if 'markdown' in self.cov_report:
+            output = self.cov_report['markdown']
+            with _backup(self.cov, 'config'):
+                with Path(output).open('w') as output_file:
+                    total = self.cov.report(ignore_errors=True, file=output_file, output_format='markdown')
+            stream.write(f'Coverage Markdown information written to file {output}\n')
+
+        # Produce Markdown report if wanted, appending to output file
+        if 'markdown-append' in self.cov_report:
+            output = self.cov_report['markdown-append']
+            with _backup(self.cov, 'config'):
+                with Path(output).open('a') as output_file:
+                    total = self.cov.report(ignore_errors=True, file=output_file, output_format='markdown')
+            stream.write(f'Coverage Markdown information appended to file {output}\n')
+
         # Produce lcov report if wanted.
         if 'lcov' in self.cov_report:
             output = self.cov_report['lcov']
@@ -230,13 +296,20 @@ class Central(CovController):
         self.cov = coverage.Coverage(
             source=self.cov_source,
             branch=self.cov_branch,
-            data_suffix=True,
+            data_suffix=_data_suffix('c'),
             config_file=self.cov_config,
         )
+        if self.cov.config.dynamic_context == 'test_function':
+            message = (
+                'Detected dynamic_context=test_function in coverage configuration. '
+                'This is unnecessary as this plugin provides the more complete --cov-context option.'
+            )
+            warnings.warn(CentralCovContextWarning(message), stacklevel=1)
+
         self.combining_cov = coverage.Coverage(
             source=self.cov_source,
             branch=self.cov_branch,
-            data_suffix=True,
+            data_suffix=_data_suffix('cc'),
             data_file=os.path.abspath(self.cov.config.data_file),  # noqa: PTH100
             config_file=self.cov_config,
         )
@@ -247,9 +320,12 @@ class Central(CovController):
         self.cov.start()
         self.set_env()
 
+        super().start()
+
     @_ensure_topdir
     def finish(self):
         """Stop coverage, save data to file and set the list of coverage objects to report on."""
+        super().finish()
 
         self.unset_env()
         self.cov.stop()
@@ -274,16 +350,22 @@ class DistMaster(CovController):
         self.cov = coverage.Coverage(
             source=self.cov_source,
             branch=self.cov_branch,
-            data_suffix=True,
+            data_suffix=_data_suffix('m'),
             config_file=self.cov_config,
         )
+        if self.cov.config.dynamic_context == 'test_function':
+            raise DistCovError(
+                'Detected dynamic_context=test_function in coverage configuration. '
+                'This is known to cause issues when using xdist, see: https://github.com/pytest-dev/pytest-cov/issues/604\n'
+                'It is recommended to use --cov-context instead.'
+            )
         self.cov._warn_no_data = False
         self.cov._warn_unimported_source = False
         self.cov._warn_preimported_source = False
         self.combining_cov = coverage.Coverage(
             source=self.cov_source,
             branch=self.cov_branch,
-            data_suffix=True,
+            data_suffix=_data_suffix('mc'),
             data_file=os.path.abspath(self.cov.config.data_file),  # noqa: PTH100
             config_file=self.cov_config,
         )
@@ -316,7 +398,7 @@ class DistMaster(CovController):
         # If worker is not collocated then we must save the data file
         # that it returns to us.
         if 'cov_worker_data' in output:
-            data_suffix = '%s.%s.%06d.%s' % (
+            data_suffix = '%s.%s.%06d.%s' % (  # noqa: UP031
                 socket.gethostname(),
                 os.getpid(),
                 random.randint(0, 999999),  # noqa: S311
@@ -330,7 +412,7 @@ class DistMaster(CovController):
                 data.read_fileobj(StringIO(output['cov_worker_data']))
                 cov.data.update(data)
             else:
-                data = CoverageData(no_disk=True)
+                data = CoverageData(no_disk=True, suffix='should-not-exist')
                 data.loads(output['cov_worker_data'])
                 cov.get_data().update(data)
             cov.stop()
@@ -361,7 +443,6 @@ class DistWorker(CovController):
 
     @_ensure_topdir
     def start(self):
-
         cleanup()
 
         # Determine whether we are collocated with master.
@@ -382,15 +463,20 @@ class DistWorker(CovController):
         self.cov = coverage.Coverage(
             source=self.cov_source,
             branch=self.cov_branch,
-            data_suffix=True,
+            data_suffix=_data_suffix(f'w{self.nodeid}'),
             config_file=self.cov_config,
         )
+        # Prevent workers from issuing module-not-measured type of warnings (expected for a workers to not have coverage in all the files).
+        self.cov._warn_unimported_source = False
         self.cov.start()
         self.set_env()
+        super().start()
 
     @_ensure_topdir
     def finish(self):
         """Stop coverage and send relevant info back to the master."""
+        super().finish()
+
         self.unset_env()
         self.cov.stop()
 
